import numpy as np
import pandas as pd
from tensorflow import keras
from tensorflow.keras import Sequential

from model_data_util.constant import OPTIONS, ONE_HOT_ENC, NUM_DIM


def convertModelToRawData(model, columns, num_data, num_dim=NUM_DIM):
    """
    Given a model, convert model conf into dataframe
    The columns of dataframe depends on the global variable options
    """
    df = pd.DataFrame(columns=columns)
    for i, l in enumerate(model.get_config()['layers']):
        l_name = l['class_name']
        conf = l['config']
        new_row = dict(zip(list(OPTIONS[l_name].keys()), [conf[k] for k in OPTIONS[l_name].keys()]))

        new_row['layer'] = l_name
        new_row['optimizer'] = model.optimizer.get_config()['name']
        new_row['loss'] = model.loss
        new_row['active'] = 1
        new_row['num_data'] = num_data
        out_shape = None
        if i > 0:
            assert model.layers[i - 1].name == conf['name']
            out_shape = np.array(model.layers[i - 1].output.shape)
        elif i == 0 and l_name == "InputLayer":
            out_shape = np.array(conf['batch_input_shape'])
        if out_shape is not None:
            assert out_shape.shape[0] <= num_dim
            for j in range(num_dim):
                if j < out_shape.shape[0]:
                    new_row[f'out_dim_{j}'] = out_shape[j]
                else:
                    new_row[f'out_dim_{j}'] = np.nan
        df = df.append(new_row, ignore_index=True)
    return df


def convertRawDataToModel(df):
    """
    :param df: the raw dataframe used to convert into model
    :return: model made by df, and the num of data
    """
    assert len(df.optimizer.unique()) == 1
    optimizer = df.optimizer.unique()[0]
    loss = df.loss.unique()[0]
    num_data = df.num_data.unique()[0]
    model = Sequential()
    drop_cols = ["active", "optimizer", "layer", "loss", "num_data", *[x for x in df.columns if "out_dim" in x]]
    for i in range(df.shape[0]):
        l_name = df.iloc[i].layer
        if l_name not in OPTIONS["Model"]["layer"]:
            # only add those layer was in our layer type choices
            print(f"WARNING: {l_name} layer was skip as not listed in OPTIONS")
            continue
        kwargs = df.iloc[i].drop(drop_cols).dropna().to_dict()
        model.add(getattr(keras.layers, l_name)(**kwargs))
    model.compile(optimizer=optimizer, loss=loss, metrics=['accuracy'])
    return model, num_data


def _preprocessRawData_Onehot_Helper(res, col, df, one_hot_enc):
    """
    :param res: return argument
    :param col: the column from df needs to be encoded into one hot
    :param df: the raw model dataframe
    :return: the one hot matrix
    """
    assert col in df.columns
    layer_one_hot = np.concatenate(
        [one_hot_enc[col].transform(np.reshape(l, (-1, 1))).toarray() for l in
         df[col].astype(str).values])
    layer_one_hot = layer_one_hot.astype(np.int)
    res = np.hstack((res, layer_one_hot))
    return res


def preprocessRawData(df, one_hot_enc=ONE_HOT_ENC, padding_to=64):
    """
    convert a raw model dataframe into one hot format
    """
    tmp_df = df.fillna(0)
    res = df.active.values.reshape(-1, 1).astype(np.int)
    columns = [["active"]]
    for col in ["layer", "optimizer", "loss", "activation"]:
        columns.append(["_".join([col, str(c)]) for c in one_hot_enc[col].categories_[0]])
        res = _preprocessRawData_Onehot_Helper(res, col, tmp_df, one_hot_enc)
    if 'use_bias' in df.columns:
        columns.append(['use_bias'])
        res = np.hstack((res, np.reshape(df['use_bias'].apply(lambda x: 1 if x else 0).values, (-1, 1))))
    for col in ["units", "filters", "num_data", *[x for x in df.columns if "out_dim" in x]]:
        if col in df.columns:
            columns.append([col])
            res = np.hstack((res, np.reshape(df[col].fillna(0).values, (-1, 1))))
    for col in ['strides', 'dilation_rate', 'kernel_size']:
        if col in df.columns:
            columns.append([col + "_r", col + "_c"])
            res = np.hstack(
                (res, np.reshape(np.array([list(x) if x == x else [0, 0] for x in df[col].values]).flatten(), (-1, 2))))
    columns = [x for y in columns for x in y]
    assert padding_to >= res.shape[0]
    res = np.concatenate([res, np.zeros((padding_to - res.shape[0], res.shape[1]))])

    return pd.DataFrame(res, columns=columns)



"""
To use evolutionary algo, we still need all global variable.
As the evolutionary algo need to modify the original models, what layers can be modified, how to modified are required
to be specified by the global variables """

import argparse
import pickle
import random

import numpy as np
import pandas as pd
import tensorflow as tf
from sklearn.model_selection import train_test_split
from tensorflow import keras
from tqdm import tqdm

from model_data_util.constant import OPTIONS
from model_data_util.create_tt_data.generate_tt_data import testTT
from model_data_util.create_tt_data.model_build import chooseRandomComb
from model_data_util.create_tt_data.model_data_convert import preprocessRawData, convertRawDataToModel


def percentageError(y_pred, y_test):
    """
    :return: the average percentage error between y_pred and y_test
    """
    return np.abs(np.mean((y_pred - y_test) / y_test))


def evaluateError(y_pred, y_test, metrix=percentageError):
    """
    :param metrix: error metrix, which can be customized
    :return: error between the actual and the prediction
    """
    y_pred = np.reshape(np.array(y_pred), (-1,))
    y_test = np.reshape(np.array(y_test), (-1,))
    return metrix(y_pred, y_test)


def evalChild(df, to):
    """
    :param df: the raw dataframe converted from model :param to: the strength of modification. "s": modify a model to
    a similar one; "d": modify a model to a different one; "dd": modify a model to a rather different one.
    :return: a new model dataframe after modification.
    """
    res_df: pd.DataFrame = df.copy()
    if to == "s":  # s stand for similar
        res_df = layerMutation(res_df)
    elif to == "d":  # d stand for slightly different
        res_df = layerMutation(res_df)
        res_df = layerAddition(res_df)
        res_df = layerRemoval(res_df)
    elif to == "dd":
        res_df = evalChild(evalChild(df, "d"), "d")
    return res_df


def layerAddition(df, max_add_layers=4, verbose=False):
    """
    :param df: raw model dataframe
    :param max_add_layers: max number of layer that will be added. The newly added row will duplicate the layer of the previous one but with different configuration.
    :param verbose: print the added rows if true
    :return: a new model dataframe after addition
    """

    # only remove the layer from part_conv
    def insert_row(idx, df, new_row):
        dfA = df.iloc[:idx, ]
        dfB = df.iloc[idx:, ]
        df = dfA.append(new_row, ignore_index=True).append(dfB).reset_index(drop=True)
        return df

    new_df = df.copy()
    max_add_layers = max(min(max_add_layers, new_df.shape[0] - 2), 1)
    # exclude the first and last layer
    add_layers = random.sample(range(1, max_add_layers + 1), 1)[0]
    selected_layers = random.sample([i for i in range(df.shape[0]) if
                                     df.iloc[i].layer in OPTIONS["Model"]["layer"]], add_layers)
    selected_layers = np.array(sorted(selected_layers))
    selected_layers += np.arange(0, selected_layers.shape[0], 1)
    if verbose:
        added_row = pd.DataFrame()
    for l in selected_layers:
        row = new_df.iloc[l]
        new_row = chooseRandomComb(OPTIONS[row.layer])
        not_mu_layer = ["optimizer", "loss"]
        for i in new_row.keys():
            if i in not_mu_layer:
                new_row[i] = row[i]
        for k in row.index:
            if k not in new_row:
                new_row[k] = row[k]
        fixed_layer = {"strides": (1, 1), "padding": "same"}
        for k in fixed_layer.keys():
            if row[k] == row[k]:
                new_row[k] = fixed_layer[k]
        if verbose:
            added_row = added_row.append(new_row, ignore_index=True)
        new_df = insert_row(l, new_df, new_row)
    if verbose:
        print("Adding Row:")
        print(added_row[df.columns])
    return new_df


def layerRemoval(df, max_remv_layers=1, removable=["Conv2D", "MaxPooling2D", "Dropout"],
                 verbose=False):
    """
    :param df: the raw model df
    :param max_remv_layers: the max number of removal layers
    :param removable: the types of layer that are removable
    :param verbose: print the removed row if True
    :return: a new model dataframe after layer removal
    """
    # only remove the layer from part_conv
    new_df = df.copy()
    max_remv_layers = max(min(max_remv_layers, new_df.shape[0] - 2), 1)
    # exclude the first and last layer
    remv_layers = random.sample(range(1, max_remv_layers + 1), 1)[0]
    selected_layers = random.sample([i for i in range(df.shape[0]) if df.iloc[i].layer in removable], remv_layers)
    if verbose:
        print("Removing Row:")
        print(new_df.iloc[selected_layers])
    new_df = new_df.drop(index=selected_layers)
    return new_df


def layerMutation(df, max_mu_layers=50, verbose=False):
    """
    This method will modify the model dataframe by randomly change the configuration of some layers
    :param df: the raw model dataframe
    :param max_mu_layers: the maximum number of mutable layers
    :return:
    """
    new_df = df.copy()
    max_mu_layers = max(min(max_mu_layers, new_df.shape[0] - 2), 1)
    # exclude the first and last layer
    mu_layers = random.sample(range(1, max_mu_layers + 1), 1)[0]
    selected_layers = random.sample(list(range(new_df.shape[0]))[1:-1], mu_layers)
    if verbose:
        print("Before Mutation")
        print(new_df.iloc[selected_layers])
    for l in selected_layers:
        row = new_df.iloc[l]
        new_row = chooseRandomComb(OPTIONS[row.layer])
        not_mu_layer = []
        if row.layer in ["Conv2D", "MaxPooling2D"]:
            if row.padding == "same":
                not_mu_layer = ["strides", "padding", "optimizer", "loss"]
            else:
                not_mu_layer = ["strides", "padding", "optimizer", "loss", "kernel_size", "pool_size"]
        for i in new_row.keys():
            if i not in not_mu_layer:
                row[i] = new_row[i]
    if verbose:
        print("After Mutation")
        print(new_df.iloc[selected_layers])
    return new_df


def modelModifier(df, error, threshold={"low": 0.1, "high": 0.25}):
    """
    :param df: the raw model dataframe
    :param error: the error of tt_predictor
    :param threshold: the threshold used to  judge similar, different, and rather different models
    :return:
    """
    new_df = df.copy()
    if error >= threshold["high"]:
        new_df = evalChild(new_df, to="s")
    elif threshold["high"] > error >= threshold["low"]:
        new_df = evalChild(new_df, to="d")
    elif error < threshold["low"]:
        new_df = evalChild(new_df, to="dd")
    return new_df


def evalChildren(X_df_list, y_list, tt_predictor):
    """
    :param X_df_list: the test model dataframe
    :param y_list: the tt of the test models
    :param tt_predictor: the model used to predict the training time of models
    :return: a set of new data from evolution
    """
    new_X_df_list: list = []
    new_y_list: list = []
    for i, X_df in enumerate(tqdm(X_df_list)):
        x = preprocessRawData(X_df).values
        x = x.reshape((-1, *x.shape))
        y = y_list[i]
        y_pred = tt_predictor(x).numpy().reshape(-1)[0]
        error = evaluateError(y_pred, y)
        new_X_df = modelModifier(X_df, error)
        new_X_df_list.append(new_X_df)
        model, _ = convertRawDataToModel(new_X_df)
        test_res = testTT(model)
        new_y_list.append(test_res['median'])
    return new_X_df_list, new_y_list


def concateModelData(X_df_list, y_list):
    """
    :param X_df_list: a list of raw model dataframe
    :param y_list: a list of tt
    :return: the preprocessed model data and tt
    """
    X = np.array([])
    y = np.array([])
    for i in range(len(X_df_list)):
        X_df = X_df_list[i]
        X_tmp = preprocessRawData(X_df).values
        X = X.reshape((-1, *np.array(X_tmp).shape))
        X_tmp = X_tmp.reshape((-1, *np.array(X_tmp).shape))
        X = np.concatenate((X, X_tmp))
    y = np.reshape(y_list, (-1, *np.array(y_list).shape[1:]))
    return X, y


def refitModel(tt_predictor, X_df_list, y_list, validation_split=0.1,
               batch_size=4, epochs=30, verbose=False):
    """
    :param tt_predictor: the predictor used to predict tt
    :param X_df_list: a list of raw model dataframe
    :param y_list: a list of tt
    :param verbose, validation_split, batch_size, epochs: the configurations while refit new data made by evolutionary process
    :return: the newly trained tt_predictor
    """
    X, y = concateModelData(X_df_list, y_list)
    tt_predictor.fit(X, y, validation_split=validation_split, epochs=epochs,
                     batch_size=batch_size, verbose=verbose)
    return tt_predictor


def evolStart(tt_predictor, num_evol, X_df_list, y_list, size_per_evol=None,
              verbose=True, disable_evol=False):
    """
    :param tt_predictor: the predictor used to predict tt
    :param num_evol: the number of evolution will experience
    :param X_df_list: a list of raw model dataframe
    :param y_list: a list of tt
    :param size_per_evol: the number of data points per evolution
    :param disable_evol: when disable evolution, the tt_predictor will be repetitively fed by the old test data
    :return: newly trained tt_predictor, the error history, the new model dataframes and corresponding tt
    """
    tt_predictor_new = keras.models.clone_model(tt_predictor)
    tt_predictor_new.compile(loss=tt_predictor.loss, optimizer=tt_predictor.optimizer.get_config()["name"])
    tt_predictor_new.set_weights(tt_predictor.get_weights())

    if size_per_evol is None:
        size_per_evol = len(X_df_list)
    assert size_per_evol <= len(X_df_list)
    error_his = []

    X_origin, y_origin = concateModelData(X_df_list, y_list)
    y_pred = tt_predictor_new.predict(X_origin)
    error = evaluateError(y_pred, y_origin)
    error_his.append(error)
    if verbose:
        print("Generation 0: error=%.3f" % error)

    index = random.sample(range(size_per_evol), size_per_evol)
    X_df_list_sample = [X_df_list[i] for i in index]
    y_list_sample = [y_list[i] for i in index]
    new_X_df_list_final = []
    new_y_list_final = []
    for i in tqdm(range(num_evol)):
        if not disable_evol:
            new_X_df_list, new_y_list = evalChildren(X_df_list_sample, y_list_sample, tt_predictor)
        else:
            new_X_df_list, new_y_list = X_df_list_sample.copy(), y_list_sample.copy()
        new_X_df_list_final.append(new_X_df_list)
        new_y_list_final.append(new_y_list)
        tt_predictor_new = refitModel(tt_predictor_new, new_X_df_list, new_y_list)
        y_pred = tt_predictor_new(X_origin)
        error = evaluateError(y_pred, y_origin)
        error_his.append(error)
        if verbose:
            print("Generation %d: error=%.3f" % (i + 1, error))
    return tt_predictor_new, error_his, new_X_df_list_final, new_y_list_final


def readData(data_path):
    tmp_res = pickle.load(open(data_path, "rb"))
    return tmp_res['X_df'], tmp_res['y']


if __name__ == "__main__":
    parser = argparse.ArgumentParser()
    parser.add_argument("--num_evol", type=int, required=True)
    parser.add_argument("--model_path", type=str, required=True)
    parser.add_argument("--data_path", type=str, required=True)
    args = parser.parse_args()
    tt_predictor = tf.keras.models.load_model(args.model_path)
    X_df_list, y_list = readData(args.data_path)
    _, X_test, _, y_test = train_test_split(X_df_list, y_list)

    tt_predictor_new, error_his_evol, new_X_df_list_final, new_y_list_final = \
        evolStart(tt_predictor, args.num_evol, X_test, y_test, disable_evol=False)



"""
# Convolutional Layer:
## Rule 1: No Convolutional Layer After the First Dense Layer
## Rule 2: At least the half of layers are convolutional
## Rule 3: Convolutional layer is always the first one
#Maxpooling Layer
## Rule 3: No Maxpooling Layer After the First Dense Layer
## Rule 4: No two Maxpooling Layers Next to Each Other
## Rule 5: Maxpooling Layers are only after convolutional layers
#Dropout Layer
## Rule 6: No two Dropout Layers Next to Each Other
## Rule 7: No dropout rate higher than 0.5
## Rule 8: Maximum two Dropout Layers
#Flatten Layer and AveragePooling Layer
## Rule 9: All CNN end with one Flatten or AveragePooling Layer followed by maximum 2 dense layer

A CNN was divided into parts:
1. convolutional_part: composed by conv2d, maxpooling2d, dropout layers
2. flatten part: composed by a flatten layer
3. dense part: composed by dense and dropout layer
4. output part: compose by a dense layer with softmax as activation
"""
import random

import numpy as np


class CnnRules:
    def __init__(self, max_layers=64):
        self.first_dense_occurred = False  # Rule: No Convolutional Layer After the First Dense Layer
        self.max_dropout = 2
        self.remain_dropout = 2
        self.initial_layer = ["Conv2D"]
        self.max_layers = max_layers  # the total max layers
        self.part_layers_num = self.generatePartLayers()
        self.layer_order = []
        self.generatePartAll()

    def generatePartLayers(self):
        """
        :return: the number of layers per parts
        """
        part_layers_num = {}
        part_layers_num["output"] = 1
        part_layers_num["faltten"] = 1
        part_layers_num["dense"] = random.choice(range(3))
        remain_max_layers = self.max_layers - np.sum(list(part_layers_num.values()))
        part_layers_num["conv"] = random.choice(range(3, remain_max_layers))
        return part_layers_num

    def generatePartConv(self):
        """
        Generate the layer_order of part_conv
        The available layers and their probabilities to occur are hard coded
        """

        def nextAvailableLayer(l_name):
            if l_name == "Conv2D":
                next_available_layer_weights = {"Conv2D": random.uniform(0.5, 1)}
                next_available_layer_weights["MaxPooling2D"] = (1 - next_available_layer_weights[
                    "Conv2D"]) * random.uniform(0.8, 1)
                next_available_layer_weights["Dropout"] = (1 - np.sum(list(next_available_layer_weights.values())))
                l = list(next_available_layer_weights.keys())
                l_w = list(next_available_layer_weights.values())
                next_layer = random.choices(l, weights=l_w)[0]
            elif l_name == "MaxPooling2D":
                # Rule: No two Maxpooling Layers Next to Each Other
                next_available_layer_weights = {"Conv2D": random.uniform(0.9, 1)}
                next_available_layer_weights["Dropout"] = (1 - np.sum(list(next_available_layer_weights.values())))
                l = list(next_available_layer_weights.keys())
                l_w = list(next_available_layer_weights.values())
                next_layer = random.choices(l, weights=l_w)[0]
            elif l_name == "Dropout":
                # Rule: the layer after dropout should only be Conv2D
                next_layer = "Conv2D"
            else:
                print("Error: received unsupported layer name: %s" % l_name)
            return next_layer

        self.layer_order = self.initial_layer.copy()
        for _ in range(len(self.layer_order), self.part_layers_num["conv"]):
            self.layer_order.append(nextAvailableLayer(self.layer_order[-1]))

    def generatePartFlatten(self):
        """
        Generate the layer_order of part_flatten
        The available layers and their probabilities to occur are hard coded
        """
        # self.layer_order.append(random.choice(["Flatten", "AveragePooling2D"]))
        self.layer_order.append("Flatten")

    def generatePartDense(self):
        """
        Generate the layer order of part_dense
        """
        if self.part_layers_num["dense"] == 1:
            self.layer_order.append(random.choice(["Dense", "Dropout"]))
        if self.part_layers_num["dense"] == 2:
            self.layer_order.append("Dense")
            self.layer_order.append("Dropout")

    def generatePartOutput(self):
        """
        Generate the layer order of part_output
        """
        self.layer_order.append("Dense")

    def generatePartAll(self):
        """
        Run all generatePartXXX functions
        """
        self.generatePartConv()
        self.generatePartFlatten()
        self.generatePartDense()
        self.generatePartOutput()



"""
The only rule of FFNN is that is was all made up by dense layers
"""

import random


class FFnnRules:
    def __init__(self, max_layers=64):
        self.max_layers = max_layers
        self.layer_order = []
        self.generateLayerOrder()

    def generateLayerOrder(self):
        self.layer_order = random.choice(range(1, self.max_layers + 1)) * ["Dense"]



import collections

import numpy as np
from sklearn.preprocessing import OneHotEncoder

# options: save changable configurations for the model
OPTIONS = collections.defaultdict(dict)

OPTIONS["Data"]["num_data"] = np.arange(1, 30, 1) * 32

OPTIONS["Model"]["layer"] = ["Conv2D", "Dense", "MaxPooling2D", "Dropout",
                             "Flatten"]  # the model's layer can be either Conv2D or Dense
OPTIONS["Model"]["optimizer"] = ["SGD", "RMSprop", "Adam", "Adadelta", "Adagrad", "Adamax", "Nadam", "Ftrl"]
OPTIONS["Model"]["loss"] = ["categorical_crossentropy", "categorical_hinge", ]

OPTIONS["Dense"]["units"] = range(1, 500)
OPTIONS["Dense"]["activation"] = ["linear", "relu", "sigmoid", "softmax", "softplus", "softsign", "tanh", "selu", "elu",
                                  "exponential"]
OPTIONS["Dense"]["use_bias"] = [True, False]

OPTIONS["Conv2D"]["filters"] = range(1, 100)
OPTIONS["Conv2D"]["kernel_size"] = list(zip(range(1, 50), range(1, 50)))  # tried product, but then RAM is in shortage
OPTIONS["Conv2D"]["strides"] = [(1, 1)] * 10 + list(zip(range(1, 3), range(1, 3)))
OPTIONS["Conv2D"]["padding"] = [*list(["same"] * 5), "valid"]
# options["Conv2D"]["dilation_rate"] = [1, 2]
OPTIONS["Conv2D"]["activation"] = ["linear", "relu", "sigmoid", "softmax", "softplus", "softsign", "tanh", "selu",
                                   "elu", "exponential"]
OPTIONS["Conv2D"]["use_bias"] = [True, False]

OPTIONS["MaxPooling2D"]["pool_size"] = list(zip(range(2, 10), range(2, 10)))
OPTIONS["MaxPooling2D"]["strides"] = [(1, 1)] * 10 + list(zip(range(1, 3), range(1, 3)))
OPTIONS["MaxPooling2D"]["padding"] = [*list(["same"] * 5), "valid"]

OPTIONS["Dropout"]["rate"] = list(np.arange(0, 0.5, 0.05))

NUM_DIM = 4  # the number of data dimension will be included in training data
dim_cols = [f"out_dim_{i}" for i in range(NUM_DIM)]

# columns: the columns name when convert model to a raw dataframe
tmp_column_set = set()
for k in OPTIONS.keys():
    if k == "Model":
        continue
    tmp_column_set = tmp_column_set.union(set(OPTIONS[k].keys()))
COLUMNS = ["active", "layer", "optimizer", "loss", *tmp_column_set, *dim_cols]

# one_hot_enc: the one hot encoder used to convert raw model dataframe into one hot
ONE_HOT_ENC = dict()
ONE_HOT_ENC["layer"] = OneHotEncoder(handle_unknown='ignore').fit(
    np.reshape(OPTIONS["Model"]["layer"], (-1, 1)))
ONE_HOT_ENC["optimizer"] = OneHotEncoder(handle_unknown='ignore').fit(
    np.reshape(OPTIONS["Model"]["optimizer"], (-1, 1)))
ONE_HOT_ENC["loss"] = OneHotEncoder(handle_unknown='ignore').fit(np.reshape(OPTIONS["Model"]["loss"], (-1, 1)))
ONE_HOT_ENC["activation"] = OneHotEncoder(handle_unknown='ignore').fit(
    np.reshape(list(set(OPTIONS["Dense"]["activation"]).union(set(OPTIONS["Conv2D"]["activation"]))), (-1, 1)))

# data_points: the number of data points used as the input to test model TT
DATA_POINTS = 32
# batch_size: the batch size used to test model TT
BATCH_SIZE = 4
# default_input_shape: the default input shape will be used when input needs to be created according to model's size
DEFAULT_INPUT_SHAPE = (32, 32, 3)

if __name__ == "__main__":
    print(f"batch_size={BATCH_SIZE}\n"
          f"default_input_shape={DEFAULT_INPUT_SHAPE}\n"
          f"set up constant options={OPTIONS}\n"
          f"set up one_hot_enc={ONE_HOT_ENC}\n"
          f"set up raw data columns={COLUMNS}")



import random

from tensorflow.keras import Sequential
from tensorflow.keras.layers import Dense, Conv2D, Flatten, MaxPooling2D, Dropout

from model_data_util.constant import DEFAULT_INPUT_SHAPE


def buildCnnModel(kwargs_list, layer_orders, out_dim):
    '''
    convert a kwargs into a cnn model
    kwargs_list and layer_orders should have the same length
    '''
    cnn = Sequential()
    for i, l in enumerate(layer_orders):
        kwargs = kwargs_list[i]
        if l == "Dense":
            cnn.add(Dense(**kwargs))
        elif l == "Conv2D":
            cnn.add(Conv2D(**kwargs))
        elif l == "MaxPooling2D":
            cnn.add(MaxPooling2D(**kwargs))
        elif l == "Dropout":
            cnn.add(Dropout(**kwargs))
        elif l == "Flatten":
            cnn.add(Flatten())
    cnn.add(Dense(out_dim, activation='softmax'))
    cnn.compile(loss=kwargs_list[-1]["loss"], optimizer=kwargs_list[-1]["optimizer"], metrics=['accuracy'])
    return cnn


def buildFFnnModel(kwargs_list, layer_orders):
    '''
    convert a kwargs into a ffnn model
    kwargs_list and layer_orders should have the same length
    '''
    ffnn = Sequential()
    for i, l in enumerate(layer_orders):
        kwargs = kwargs_list[i]
        if l == "Dense":
            ffnn.add(Dense(**kwargs))
        elif l == "Dropout":
            ffnn.add(Dropout(**kwargs))
    ffnn.compile(loss=kwargs_list[-1]["loss"], optimizer=kwargs_list[-1]["optimizer"], metrics=['accuracy'])
    return ffnn


def chooseRandomComb(options_layer):
    res = dict()
    for k, v in options_layer.items():
        res[k] = (random.sample(v, 1)[0])
    return res


def generateRandomModelConfigList(layer_orders, options, max_num_hidden_layer=62, input_shape=DEFAULT_INPUT_SHAPE):
    '''
    Use gloabl variable all_comb to generate random cnn model conf
    To build a model, pass the return to buildCnnModel method
    '''

    def updateImageShape(l, kwargs, image_shape):
        if l == "Conv2D":
            if type(kwargs["kernel_size"]) == int:  # when kwargs["kernel_size"] was set by int
                kernel_size = (kwargs["kernel_size"], kwargs["kernel_size"])
            else:
                kernel_size = kwargs["kernel_size"]
        elif l == "MaxPooling2D":
            if type(kwargs["pool_size"]) == int:  # when kwargs["kernel_size"] was set by int
                # for program simplicity, I called pool_size as kernel_size
                kernel_size = (kwargs["pool_size"], kwargs["pool_size"])
            else:
                kernel_size = kwargs["pool_size"]

        if type(kwargs["strides"]) == int:  # when kwargs["strides"] was set by int
            strides = (kwargs["strides"], kwargs["strides"])
        else:
            strides = kwargs["strides"]
        if kwargs["padding"] == "valid":
            image_shape[0] = (image_shape[0] - kernel_size[0]) // strides[0] + 1
            image_shape[1] = (image_shape[1] - kernel_size[1]) // strides[1] + 1
        if kwargs["padding"] == "same":
            if image_shape[0] % strides[0] > 0:
                image_shape[0] = image_shape[0] // strides[0] + 1
            else:
                image_shape[0] = image_shape[0] // strides[0]
            if image_shape[1] % strides[1] > 0:
                image_shape[1] = image_shape[1] // strides[1] + 1
            else:
                image_shape[1] = image_shape[1] // strides[1]
        assert image_shape[0] > 0 and image_shape[1] > 0
        return image_shape

    def validKernelStridesSize(l, kwargs, image_shape):
        if l == "Conv2D":
            if type(kwargs["kernel_size"]) == int:
                kernel_size = (kwargs["kernel_size"], kwargs["kernel_size"])
            else:
                kernel_size = kwargs["kernel_size"]
        elif l == "MaxPooling2D":
            if type(kwargs["pool_size"]) == int:  # when kwargs["kernel_size"] was set by int
                # for program simplicity, I called pool_size as kernel_size
                kernel_size = (kwargs["pool_size"], kwargs["pool_size"])
            else:
                kernel_size = kwargs["pool_size"]

        if type(kwargs["strides"]) == int:
            strides = (kwargs["strides"], kwargs["strides"])
        else:
            strides = kwargs["strides"]
        judge = True
        if l in ["Conv2D", "MaxPooling2D"]:
            judge = judge and (kernel_size[0] <= image_shape[0] and kernel_size[1] <= image_shape[1])
        judge = judge and (strides[0] <= image_shape[0] and strides[1] <= image_shape[1])
        if judge:
            return True
        else:
            return False

    kwargs_list = []
    image_shape: list = list(input_shape[:2])
    image_shape_list: list = []
    # image_shape should end up in the same shape as model
    new_layer_orders = []
    max_strides = [3, 3]
    for i, l in enumerate(layer_orders):
        if l == "Dense":
            kwargs = chooseRandomComb(options["Dense"])
        elif l == "Conv2D":
            if (image_shape[0] == 1 or image_shape[1] == 1):
                # if one of the image dim has only size one, we stop adding new conv2D
                continue
            options_conv2d = options["Conv2D"].copy()
            # always ensure the kernel and strides size is smaller than the image
            options_conv2d["kernel_size"] = list(zip(range(1, image_shape[0]), range(1, image_shape[1])))
            options_conv2d["strides"] = [(1, 1)] * 10 + list(zip(range(1, max_strides[0]),
                                                                 range(1, max_strides[1])))
            kwargs = chooseRandomComb(options_conv2d)
            image_shape = updateImageShape(l, kwargs, image_shape)
            max_strides = [min(max_strides[0], max(1, image_shape[0])), min(max_strides[1], max(1, image_shape[1]))]
        elif l == "MaxPooling2D":
            if (image_shape[0] == 1 or image_shape[1] == 1):
                # if one of the image dim has only size one, we stop adding new conv2D
                continue
            options_maxpooling2d = options["MaxPooling2D"].copy()
            options_maxpooling2d["pool_size"] = list(zip(range(1, image_shape[0]), range(1, image_shape[1])))
            options_maxpooling2d["strides"] = [(1, 1)] * 10 + list(zip(range(1, max_strides[0]),
                                                                       range(1, max_strides[1])))
            kwargs = chooseRandomComb(options_maxpooling2d)
            image_shape = updateImageShape(l, kwargs, image_shape)
            max_strides = [min(max_strides[0], max(1, image_shape[0])), min(max_strides[1], max(1, image_shape[1]))]
        elif l == "Dropout":
            kwargs = chooseRandomComb(options["Dropout"])
        elif l == "Flatten":
            kwargs = {}
        # elif l == "AveragePooling2D":
        #   pass
        else:
            print("Error: layer order contained unsupported layer: %s" % l)
        kwargs_list.append(kwargs)
        new_layer_orders.append(l)
        image_shape_list.append(image_shape.copy())
    kwargs_list.append({"optimizer": random.sample(options['Model']['optimizer'], 1)[0],
                        "loss": random.sample(options['Model']['loss'], 1)[0]})
    return kwargs_list, new_layer_orders, image_shape_list



import argparse
import collections
import random
import time

import numpy as np
from tensorflow import keras
from tqdm import tqdm

from model_data_util.constant import DEFAULT_INPUT_SHAPE, OPTIONS, COLUMNS
from model_data_util.create_tt_data.cnn_build_rule import CnnRules
from model_data_util.create_tt_data.ffnn_build_rule import FFnnRules
from model_data_util.create_tt_data.model_build import buildCnnModel, generateRandomModelConfigList, buildFFnnModel
from model_data_util.create_tt_data.model_data_convert import convertModelToRawData


class TimeHistory(keras.callbacks.Callback):
    def on_train_begin(self, logs={}):
        self.times = []

    def on_epoch_begin(self, batch, logs={}):
        self.epoch_time_start = time.time()

    def on_epoch_end(self, batch, logs={}):
        self.times.append(time.time() - self.epoch_time_start)


def createInputbyModel(model, data_points, data_shape=DEFAULT_INPUT_SHAPE):
    """
    Generate data according to model's conf
    The input shape is set in default
    """
    last_dense_conf = {}
    first_layer_conf = {}
    for l_conf in model.get_config()['layers']:
        if 'Dense' == l_conf['class_name']:
            last_dense_conf = l_conf['config']
        if first_layer_conf == {} and 'InputLayer' != l_conf['class_name']:
            first_layer_conf = l_conf
    assert last_dense_conf != {}
    out_shape = last_dense_conf['units']
    if first_layer_conf['class_name'] == 'Dense':
        data_shape = [out_shape] if type(out_shape) == int else out_shape
    x = np.ones((data_points, *data_shape))
    y = np.ones((data_points, out_shape))
    return (x, y)


def testTT(model, epochs=7, batch_size=4):
    """
    Given a model, test its TT. The first epoch is excluded to avoid setup time
    """
    num_data = random.choice(OPTIONS["Data"]["num_data"])
    x, y = createInputbyModel(model, num_data)
    time_callback = TimeHistory()
    model.fit(x, y, epochs=epochs, batch_size=batch_size, callbacks=[time_callback], verbose=False)
    times = time_callback.times
    test_res = {}
    test_res['median'] = np.median(times[2:])
    test_res['mean'] = np.mean(times[2:])
    test_res['std'] = np.std(times[2:])
    test_res['x_shape'] = np.array(x).shape
    test_res['y_shape'] = np.array(y).shape
    return test_res


def generateTTData(num_data, out_dim=10, max_layers=32, type="cnn", options=OPTIONS, columns=COLUMNS):
    result = collections.defaultdict(list)

    for _ in tqdm(range(num_data)):
        rules = None
        if type == "cnn":
            rules = CnnRules(max_layers=max_layers)
        elif type == "ffnn":
            rules = FFnnRules(max_layers=max_layers)
        assert rules is not None
        kwargs_list, layer_orders, image_shape_list = generateRandomModelConfigList(rules.layer_order, options=options)
        if type == "cnn":
            model = buildCnnModel(kwargs_list, layer_orders, out_dim)
        elif type == "ffnn":
            model = buildFFnnModel(kwargs_list, layer_orders)
        test_res = testTT(model)
        df = convertModelToRawData(model, columns, test_res['x_shape'][0])
        result["X_df"].append(df)
        result["y_median"].append(test_res['median'])
        result["y_mean"].append(test_res['mean'])
        result["y_std"].append(test_res['std'])
        if type == "cnn":
            result["image_shape_list"].append(image_shape_list)
    return result


if __name__ == "__main__":
    parser = argparse.ArgumentParser()
    parser.add_argument('--num_data', type=int, required=True)
    parser.add_argument('--out_dim', type=int, required=True)
    args = parser.parse_args()
    num_data = args.num_data
    out_dim = args.out_dim

    result = generateTTData(num_data, out_dim, type="ffnn")
    # print(convertRawDataToModel(result["X_df"][0]))
    # print(preprocessRawData(result["X_df"][0]).columns)

    # pickle.dump(result, open(f"../data/local_dp{DATA_POINTS}_CNN_Data_1.pkl", 'wb'))
